/***************************************************************************
Copyright (c) 2016, Guillaume Baud-Berthier, SafeRiver/LaBRI

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
***************************************************************************/

#include "zigzag_bwd.h"

int memo_bad = -1;

std::vector<int>   act_init_bwd;
//std::vector<int>   var2cycle_bwd;
std::vector<idef*> var2idef_bwd;

// Simple path constraints on demand
bool simple_path_od_bwd(int time, bool rep, bool verbose) {
  vec<Lit> lits;
  idef* tmp_def;
  dll_elem<idef> *tmp_e;
  bool same_state;
  bool run_solver_again = 1;
  
  while (run_solver_again) {
    run_solver_again = 0;
     
    for(int i = 0; i < time && !run_solver_again; i++) {
      for(int j = i + 1; j <= time && !run_solver_again; j++) {
	
	same_state = 1;
	
	tmp_e = dll_latches->head;
	while(tmp_e != NULL && same_state) {
	  tmp_def = tmp_e->def;

	  assert(tmp_def->lhs_ts[i] > 1);
	  assert(tmp_def->lhs_ts[j] > 1);
	  assert(solver->model[tmp_def->lhs_ts[i] >> 1] != l_Undef);
	  assert(solver->model[tmp_def->lhs_ts[j] >> 1] != l_Undef);

	  if (solver->model[tmp_def->lhs_ts[i] >> 1] == l_Undef ||
	      solver->model[tmp_def->lhs_ts[j] >> 1] == l_Undef ||
	      solver->model[tmp_def->lhs_ts[i] >> 1] != solver->model[tmp_def->lhs_ts[j] >> 1] ||
	      ((tmp_def->lhs_ts[i] >> 1 == tmp_def->lhs_ts[j] >> 1) &&
	       (tmp_def->lhs_ts[i] >> 1 != tmp_def->lhs_ts[j])) )
	    same_state = 0;
	  
	  tmp_e = tmp_e->next;
	}
	
	if (same_state) {
	  // Add constraints
	  lits.clear();
	  tmp_e = dll_latches->head;
	  while(tmp_e != NULL && same_state) {
	    tmp_def = tmp_e->def;

	    assert(tmp_def->lhs_ts[i] > 1);
	    assert(tmp_def->lhs_ts[j] > 1);
	    
	    unsigned new_v = new_var() >> 1;
	    //var2cycle_bwd.push_back(-2);
	    solver->var2cycle_t.push();
	    solver->var2cycle_t.last() = -2;

	    var2idef_bwd.push_back(NULL);

	    //assert(var2cycle_bwd[new_v] == -2);
	    assert(solver->var2cycle_t[new_v] == -2);
	    assert(var2idef_bwd[new_v] == NULL);

	    while (new_v >= solver->nVars()) solver->newVar();

	    assert(aiger_sign(tmp_def->lhs_ts[i]) == aiger_sign(tmp_def->lhs_ts[j]));
	    lits.push(mkLit(new_v));

	    tcl(-new_v, -(tmp_def->lhs_ts[i] >> 1), -(tmp_def->lhs_ts[j] >> 1));
	    tcl(-new_v,  (tmp_def->lhs_ts[i] >> 1),  (tmp_def->lhs_ts[j] >> 1));
	    tcl( new_v,  (tmp_def->lhs_ts[i] >> 1), -(tmp_def->lhs_ts[j] >> 1));
	    tcl( new_v, -(tmp_def->lhs_ts[i] >> 1),  (tmp_def->lhs_ts[j] >> 1));

	    tmp_e = tmp_e->next;
	  }
	  
	  solver->addClause_(lits);
	  run_solver_again = 1;
	}
      }
    }

    if (run_solver_again) {
      if (rep)
	//assume(solver->bad_lit);
	assume(memo_bad);
      if (check_sat(verbose) == l_False)
	return 1;
    }

  }

  return 0;
}


dll_list<ands_def>* unroll_kind_bwd(int time) {
  dll_list<ands_def> *tmp_list_ands;
  tmp_list_ands = new dll_list<ands_def>();
  dll_init<ands_def>(tmp_list_ands);
  
  idef* tmp_def;
  dll_elem<idef> *tmp_e;

  // Latches
  tmp_e = dll_latches->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;
    
    tmp_def->lhs_ts.push_back(new_var());
    //var2cycle_bwd.push_back(time);
    solver->var2cycle_t.push();
    solver->var2cycle_t.last() = time;

    var2idef_bwd.push_back(tmp_def);

    bcl(-(tmp_def->lhs_ts[time] >> 1), -act_init_bwd[time]);

    tmp_e = tmp_e->next;
  }

  // Inputs
  tmp_e = dll_inputs->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;

    tmp_def->lhs_ts.push_back(new_var());
    //var2cycle_bwd.push_back(time);
    solver->var2cycle_t.push();
    solver->var2cycle_t.last() = time;

    var2idef_bwd.push_back(tmp_def);

    tmp_e = tmp_e->next;
  }


  // Ands
  tmp_e = dll_ands->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;

    prop_tr_un *tpu = unroll_prop_tr(tmp_def->rhs_tr, time, 0);

    assert(tpu->lit == UNDEF_LIT);
    assert(tpu->op != VAR);

    tmp_def->lhs_ts.push_back(new_var());
    //var2cycle_bwd.push_back(time);
    solver->var2cycle_t.push();
    solver->var2cycle_t.last() = time;

    var2idef_bwd.push_back(tmp_def);

    ands_def *tmp_ands = new ands_def(tmp_def->lhs_ts[time], tpu, tmp_def, time);

    map_ands.insert(std::pair<unsigned, ands_def*>(aiger_lit2var(tmp_ands->lhs), tmp_ands));
    
    tmp_ands->dllade = dll_insert_last(tmp_list_ands, aiger_lit2var(tmp_def->lhs_ts[time]), tmp_ands);

    tmp_e = tmp_e->next;
  }

  // Add relation constraints
  tmp_e = dll_latches->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;
    
    int lit_latch = tmp_def->lhs_ts[time-1];
    if (tmp_def->rhs.front() > 1) {
      int lit_def   = mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time];
      if ((!aiger_sign(lit_latch) && !aiger_sign(lit_def)) ||
	  ( aiger_sign(lit_latch) &&  aiger_sign(lit_def))) {
	if (aiger_sign(tmp_def->rhs.front())) {
	  bcl(  lit_latch >> 1,    lit_def >> 1);
	  bcl(-(lit_latch >> 1), -(lit_def >> 1));
	} else {
	  bcl(  lit_latch >> 1,  -(lit_def >> 1));
	  bcl(-(lit_latch >> 1),   lit_def >> 1);
	}
      } else {
	if (aiger_sign(tmp_def->rhs.front())) {
	  bcl(  lit_latch >> 1,  -(lit_def >> 1));
	  bcl(-(lit_latch >> 1),   lit_def >> 1);
	} else {
	  bcl(  lit_latch >> 1,    lit_def >> 1);
	  bcl(-(lit_latch >> 1), -(lit_def >> 1));
	}
      }
    } else {
      ucl( tmp_def->rhs.front() ? lit_latch >> 1 : -(lit_latch >> 1));
    }

    tmp_e = tmp_e->next;
  }

  return tmp_list_ands;
}


void replicate_clauses_bwd(int current_cycle, bool verbose_) {
  vec<Lit> lits;
  cl_rep cl;
  int diff_cycle;
  int stats_rep = 0;
  for(int i = 0; i < solver->cl_rep_vec.size(); i++) {
    cl = solver->cl_rep_vec[i];
    diff_cycle = current_cycle - cl.learnt_cycle;

    for (int j = 0; j < cl.size_cl; j++) {
      //int cycle_var = var2cycle_bwd[var(cl.lits[j])];
      int cycle_var = solver->var2cycle_t[var(cl.lits[j])];
      assert(cycle_var > -1);

      assert(var2idef_bwd[var(cl.lits[j])] != NULL);
      if (var2idef_bwd[var(cl.lits[j])] == (idef*)-1) {
	lits.push(mkLit(act_init_bwd[cycle_var + diff_cycle],
			sign(cl.lits[j])));
	
      } else {
	assert(var2idef_bwd[var(cl.lits[j])]->lhs_ts[cycle_var] >> 1 == var(cl.lits[j]));
	
	lits.push(mkLit(var2idef_bwd[var(cl.lits[j])]->lhs_ts[cycle_var + diff_cycle] >> 1,
			sign(cl.lits[j]) != 
			aiger_sign(var2idef_bwd[var(cl.lits[j])]->lhs_ts[cycle_var + diff_cycle])));
      }

    }
    
    if (solver->rep_in_learntDB)
      solver->addClause_learntDB(lits, cl.lbd);
    else
      solver->Solver::addClause_(lits);
    stats_rep++;
    
    lits.clear();
  }

  if (verbose_)
    if (solver->rep_max_sz < INT_MAX)
      printf("[zigzag]  %d/%d clauses (max size %d) replicated to new cycle!\n", stats_rep,
	     solver->cl_rep_vec.size(), solver->rep_max_sz);
    else
      printf("[zigzag]  %d/%d clauses replicated to new cycle!\n", stats_rep, solver->cl_rep_vec.size());
}

/**************************************************************************************/

int zigzag_bwd(int start, int stop, aiger *model, bool elim_hl, int simple, int rep, bool repG,
	       int rep_max_sz, bool verbose_) {
  int i, j;

  int status = -1;

  unsigned bad_lit = init_model(model, stop, 0, verbose_);

  
  if (elim_hl)
    hl_ve(verbose_);

  //var2cycle_bwd.push_back(-3);
  solver->var2cycle_t.push();
  solver->var2cycle_t.last() = -3;

  var2idef_bwd.push_back(NULL);

  int nb_vars_by_cycle = dll_latches->sz + dll_inputs->sz + dll_ands->sz + 1;
  
  solver->nb_vars_by_cycle = nb_vars_by_cycle;
  solver->rep_max_sz = rep ? rep_max_sz : 0;
  solver->rep_in_learntDB = rep == 2;
  solver->repGlue = repG;

  for(i = 0; i <= stop; i++) {
    solver->current_cycle = i;
 
    act_init_bwd.push_back(new_var() >> 1);
    //var2cycle_bwd.push_back(i);
    solver->var2cycle_t.push();
    solver->var2cycle_t.last() = i;

    var2idef_bwd.push_back((idef*)-1);

    if (verbose_) printf("[zigzag]  Encoding %d..\n", i);

    list_ands.push_back(unroll_kind_bwd(i));

    iter_flatten_ptu(i);

    ands2clauses(i);

    // if (rep == 1) // Replicate in original clause DB
    //   replicate_clauses_bwd(i, 0, verbose_);
    // else if (rep == 2) // Replicate in learnt clauses DB
    //   replicate_clauses_bwd(i, 1, verbose_);
    if (rep)
      replicate_clauses_bwd(i, verbose_);

    assert(mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] > 1);
    
    // Add Bad cls
    int lit_bad;
    lit_bad = mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i];
    std::map<unsigned, ands_def*>::iterator lit_bad_ref = map_ands.find(aiger_lit2var(lit_bad));
    if (lit_bad_ref != map_ands.end())
      assert(lit_bad_ref->second->cst == -1);
      
    if (aiger_sign(lit_bad))
      lit_bad = aiger_sign(bad_lit) ?  (lit_bad >> 1) : -(lit_bad >> 1);
    else
      lit_bad = aiger_sign(bad_lit) ? -(lit_bad >> 1) :  (lit_bad >> 1);

    if (i) 
      ucl(-lit_bad); // Assert:  P(x_i);
    else if (!rep)
      ucl(lit_bad);  // Assert: -P(x_0);
    else {
      //solver->bad_lit = lit_bad;
      memo_bad = lit_bad;
      //var2cycle_bwd[abs(lit_bad)] = -1;
      solver->var2cycle_t[abs(lit_bad)] = -2;
    }
    if (verbose_) printf("[zigzag]  Solving %d-induction..\n", i);

    if (rep)
      assume(memo_bad); // Assume: -P(x_0);
      //assume(solver->bad_lit); // Assume: -P(x_0);

    lbool res_check = check_sat(verbose_); 
    if (res_check == l_False) {
      status = 0;
      break;
    } else if (i > 1 && simple == 2) {
      if (simple_path_od_bwd(i, rep, verbose_)) {
	status = 0;
	break;
      }
    }

    if (verbose_) printf("[zigzag]  Solving BMC %d..\n", i);

    assume(act_init_bwd[i]);           // Assume:  I(x_i)
    if (rep)
      assume(memo_bad); // Assume: -P(x_0);
      //assume(solver->bad_lit);         // Assume: -P(x_0);

    res_check = check_sat(verbose_); 
    if (res_check == l_True) {
      status = 1;
      break;
    }


    printf("u%d\n", i);
    fflush(stdout);
  }

  printf("%d\nb0\n.\n", status);
  fflush(stdout);

  if (verbose_)
    printf("Vars %d (%d - %d)| Clauses %d\n",
	   solver->nVars() - solver->eliminated_vars, solver->nVars(), solver->eliminated_vars, solver->nClauses());
  destroy_model(verbose_);
  return 0;
}

