/***************************************************************************
Copyright (c) 2016, Guillaume Baud-Berthier, SafeRiver/LaBRI

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
***************************************************************************/

#include "zigzag_t.h"

unsigned act_init_t;
//std::vector<int> init_t;
//std::vector<int> shift_spod_t;
//std::vector<int> var2cycle_t;

// Simple path (loop free) constraints on demand
bool simple_path_od_t(int time, int bad, bool verbose_) {

  vec<Lit> lits;
  idef* tmp_def;
  dll_elem<idef> *tmp_e;
  bool same_state;
  bool run_solver_again = 1;

  while (run_solver_again) {
    run_solver_again = 0;

    for(int i = 0; i < time && !run_solver_again; i++) {
      for(int j = i+1; j <= time && !run_solver_again; j++) {
	
	same_state = 1;

	tmp_e = dll_latches->head;
	while(tmp_e != NULL && same_state) {
	  tmp_def = tmp_e->def;

	  assert(tmp_def->lhs_ts[i] > 1);
	  assert(tmp_def->lhs_ts[j] > 1);

	  if (solver->model[tmp_def->lhs_ts[i] >> 1] != solver->model[tmp_def->lhs_ts[j] >> 1])
	    same_state = 0;

	  tmp_e = tmp_e->next;
	}

	if (same_state) {
	  // Add constraints
	  lits.clear();
	  tmp_e = dll_latches->head;
	  while(tmp_e != NULL && same_state) {
	    tmp_def = tmp_e->def;


	    assert(tmp_def->lhs_ts[i] > 1);
	    assert(tmp_def->lhs_ts[j] > 1);


	    unsigned new_v = new_var() >> 1;
	    //solver->var2cycle_t.push(-2);
	    solver->var2cycle_t.push();
	    solver->var2cycle_t.last() = -2;
	    assert(solver->var2cycle_t[new_v] == -2);
	    while (new_v >= solver->nVars()) solver->newVar();

	    assert(!aiger_sign(tmp_def->lhs_ts[i]) && !aiger_sign(tmp_def->lhs_ts[j]));

	    lits.push(mkLit(new_v));

	    // _5cl(-new_v, -(tmp_def->lhs_ts[i] >> 1), -(tmp_def->lhs_ts[j] >> 1), -(solver->act_trans_t[i]), -(solver->act_trans_t[j]));
	    // _5cl(-new_v,  (tmp_def->lhs_ts[i] >> 1),  (tmp_def->lhs_ts[j] >> 1), -(solver->act_trans_t[i]), -(solver->act_trans_t[j]));
	    // _5cl( new_v,  (tmp_def->lhs_ts[i] >> 1), -(tmp_def->lhs_ts[j] >> 1), -(solver->act_trans_t[i]), -(solver->act_trans_t[j]));
	    // _5cl( new_v, -(tmp_def->lhs_ts[i] >> 1),  (tmp_def->lhs_ts[j] >> 1), -(solver->act_trans_t[i]), -(solver->act_trans_t[j]));

	    tcl(-new_v, -(tmp_def->lhs_ts[i] >> 1), -(tmp_def->lhs_ts[j] >> 1));
	    tcl(-new_v,  (tmp_def->lhs_ts[i] >> 1),  (tmp_def->lhs_ts[j] >> 1));
	    tcl( new_v,  (tmp_def->lhs_ts[i] >> 1), -(tmp_def->lhs_ts[j] >> 1));
	    tcl( new_v, -(tmp_def->lhs_ts[i] >> 1),  (tmp_def->lhs_ts[j] >> 1));


	    solver->shift_spod_t[time + 1] = solver->shift_spod_t[time + 1] + 1;

	    tmp_e = tmp_e->next;
	  }

	  // lits.push(~mkLit(solver->act_trans_t[i]));
	  // lits.push(~mkLit(solver->act_trans_t[j]));

	  solver->addClause_(lits);
	  run_solver_again = 1;
	}
      }
    }

    if (run_solver_again) {
      assume(bad); // Assume: -P(x)

      for (int k = 0; k <= time; k++)
	assume(solver->act_trans_t[k]); // Assume: T(x, x')

      if (check_sat(verbose_) == l_False)
	return 1;
    }
  }

  return 0;
}


// Simple path (loop free) constraints 
void simple_path_t(int time) {
  vec<Lit> lits;
  idef* tmp_def;
  dll_elem<idef> *tmp_e;  
  for (int i = 0; i < time; i++) {

    tmp_e = dll_latches->head;
    while(tmp_e != NULL) {
      tmp_def = tmp_e->def;

      assert(tmp_def->lhs_ts[i] > 1);
      assert(tmp_def->lhs_ts[time] > 1);

      unsigned new_v = new_var() >> 1;
      solver->var2cycle_t.push();
      solver->var2cycle_t.last() = -2;
      assert(solver->var2cycle_t[new_v] == -2);
      while (new_v >= solver->nVars()) solver->newVar();
      
      assert(!aiger_sign(tmp_def->lhs_ts[i]) && !aiger_sign(tmp_def->lhs_ts[time]));
      
      lits.push(mkLit(new_v));

      tcl(-new_v, -(tmp_def->lhs_ts[i] >> 1), -(tmp_def->lhs_ts[time] >> 1));
      tcl(-new_v,  (tmp_def->lhs_ts[i] >> 1),  (tmp_def->lhs_ts[time] >> 1));
      tcl( new_v,  (tmp_def->lhs_ts[i] >> 1), -(tmp_def->lhs_ts[time] >> 1));
      tcl( new_v, -(tmp_def->lhs_ts[i] >> 1),  (tmp_def->lhs_ts[time] >> 1));

      tmp_e = tmp_e->next;
    }
    
    solver->addClause_(lits);
    lits.clear();
  }
}


dll_list<ands_def>* unroll_kind_t(int time) {
  dll_list<ands_def> *tmp_list_ands;
  tmp_list_ands = new dll_list<ands_def>();
  dll_init<ands_def>(tmp_list_ands);
  
  idef* tmp_def;
  dll_elem<idef> *tmp_e;

  // Inputs
  tmp_e = dll_inputs->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;
    
    tmp_def->lhs_ts.push_back(new_var());
    solver->var2cycle_t.push();
    solver->var2cycle_t.last() = time;
    tmp_e = tmp_e->next;
  }  

  int it_l = 0;
  // Latches
  tmp_e = dll_latches->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;

    if (!time) { // Init
      tmp_def->lhs_ts.push_back(new_var());
      solver->var2cycle_t.push();
      solver->var2cycle_t.last() = time;
      
      // tcl( (tmp_def->lhs_ts[0] >> 1), -(init_t[it_l]), -(solver->act_trans_t[0]));
      // tcl(-(tmp_def->lhs_ts[0] >> 1),   init_t[it_l] , -(solver->act_trans_t[0]));
      // bcl(-(init_t[it_l++]), -act_init_t);

      bcl(-(tmp_def->lhs_ts[0] >> 1), -act_init_t);
      
    } else {
      assert(tmp_def->rhs.size() == 1);
      
      if (tmp_def->rhs.front() > 1) {
	assert(tmp_def->lhs_ts[time-1] != -1);

	assert(tmp_def->lhs_ts.size() == time);
	tmp_def->lhs_ts.push_back(new_var());
	solver->var2cycle_t.push();
	solver->var2cycle_t.last() = time;

	if (aiger_sign(tmp_def->rhs.front()) != aiger_sign(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1])) {
	  tcl( (tmp_def->lhs_ts[time] >> 1),  (mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1), -(solver->act_trans_t[time]));
	  tcl(-(tmp_def->lhs_ts[time] >> 1), -(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1), -(solver->act_trans_t[time]));


	  //bcl( (tmp_def->lhs_ts[time] >> 1),  (mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1));
	  //bcl(-(tmp_def->lhs_ts[time] >> 1), -(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1));

	  // tcl( (tmp_def->lhs_ts[time] >> 1),  (mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1), -(solver->act_trans_t[0]));
	  // tcl(-(tmp_def->lhs_ts[time] >> 1), -(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1), -(solver->act_trans_t[0]));


	} else {
	  tcl( (tmp_def->lhs_ts[time] >> 1), -(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1), -(solver->act_trans_t[time]));
	  tcl(-(tmp_def->lhs_ts[time] >> 1),   mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1,  -(solver->act_trans_t[time]));


	  // bcl( (tmp_def->lhs_ts[time] >> 1), -(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1));
	  // bcl(-(tmp_def->lhs_ts[time] >> 1),   mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1);

	  // tcl( (tmp_def->lhs_ts[time] >> 1), -(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1), -(solver->act_trans_t[0]));
	  // tcl(-(tmp_def->lhs_ts[time] >> 1),   mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1,  -(solver->act_trans_t[0]));

	}
	

      } else { // Cst latch
	tmp_def->lhs_ts.push_back(new_var());
	solver->var2cycle_t.push();
	solver->var2cycle_t.last() = time;
	assert(tmp_def->rhs.front() == 1);
	bcl(tmp_def->lhs_ts[time] >> 1, -(solver->act_trans_t[time]));

	//ucl(tmp_def->lhs_ts[time] >> 1);

	//bcl(tmp_def->lhs_ts[time] >> 1, -(solver->act_trans_t[0]));
      }
    }
    tmp_e = tmp_e->next;
  }
  
  // Ands
  tmp_e = dll_ands->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;

    prop_tr_un *tpu = unroll_prop_tr(tmp_def->rhs_tr, time, 0);

    assert(tpu->lit == UNDEF_LIT);
    assert(tpu->op != VAR);

    tmp_def->lhs_ts.push_back(new_var());
    solver->var2cycle_t.push();
    solver->var2cycle_t.last() = time;
    
    ands_def *tmp_ands = new ands_def(tmp_def->lhs_ts[time], tpu, tmp_def, time);

    map_ands.insert(std::pair<unsigned, ands_def*>(aiger_lit2var(tmp_ands->lhs), tmp_ands));
    
    tmp_ands->dllade = dll_insert_last(tmp_list_ands, aiger_lit2var(tmp_def->lhs_ts[time]), tmp_ands);


    tmp_e = tmp_e->next;
  }

  return tmp_list_ands;
}



void replicate_clauses_t(int current_cycle, bool verbose_) {
  vec<Lit> lits;
  cl_rep cl;
  int diff_cycle;
  int stats_rep = 0;
  for(int i = 0; i < solver->cl_rep_vec.size(); i++) {
    cl = solver->cl_rep_vec[i];
    //diff_cycle = current_cycle - cl.learnt_cycle;
    diff_cycle = current_cycle - cl.maxc - 1;
    
    int rep = 1;
    for (int j = 0; j < cl.size_cl && rep; j++) {
      int cycle_var = solver->var2cycle_t[var(cl.lits[j])];

      assert(cycle_var > -1);
      int var_wo_shift = var(cl.lits[j]) - solver->shift_spod_t[cycle_var];
      lits.push(mkLit(var_wo_shift + (solver->nb_vars_by_cycle * diff_cycle) +
		      solver->shift_spod_t[cycle_var + diff_cycle], sign(cl.lits[j])));
      
      // printf("REP (c %d) : %d (%d) -> %d (%d)\n", cycle_var, var(cl.lits[j]), var_wo_shift,
      // 	     var_wo_shift + (solver->nb_vars_by_cycle * diff_cycle) +
      // 	     solver->shift_spod_t[cycle_var + diff_cycle],
      // 	     var_wo_shift + (solver->nb_vars_by_cycle * diff_cycle));
      
    }
    
    if (solver->rep_in_learntDB)
      solver->addClause_learntDB(lits, cl.lbd);
    else
      solver->Solver::addClause_(lits);
    stats_rep++;

    lits.clear();
  }

  if (verbose_)
    if (solver->rep_max_sz < INT_MAX)
      printf("[zigzag]  %d/%d clauses (max size %d) replicated to new cycle!\n", stats_rep,
	     solver->cl_rep_vec.size(), solver->rep_max_sz);
    else
      printf("[zigzag]  %d/%d clauses replicated to new cycle!\n", stats_rep, solver->cl_rep_vec.size());

}



int zigzag_t(int start, int stop, aiger *model, bool elim_hl, int simple, int rep, bool repG,
	   int rep_max_sz, bool verbose_) {
  int i, j;

  int status = -1;

  unsigned bad_lit = init_model(model, stop, 0, verbose_);

  if (elim_hl)
    hl_ve(verbose_);

  solver->var2cycle_t.push();
  solver->var2cycle_t.last() = -3;

  act_init_t = new_var() >> 1;

  solver->var2cycle_t.push();
  solver->var2cycle_t.last() = -2;

  int nb_vars_by_cycle = dll_latches->sz + dll_inputs->sz + dll_ands->sz + 1;

  solver->nb_vars_by_cycle = nb_vars_by_cycle;
  solver->rep_max_sz = rep ? rep_max_sz : 0;
  solver->repGlue = repG;
  solver->intern_rep = 1;
  //solver->init_lit = act_init_t;
  //solver->bad_lit = act_init_t;
  solver->rep_in_learntDB = rep == 2;
  solver->shift_spod_t.push();
  solver->shift_spod_t.last() = 0;

  for(i = 0; i <= stop; i++) {
    solver->current_cycle = i;

    solver->act_trans_t.push();
    solver->act_trans_t.last() = new_var() >> 1;

    solver->var2cycle_t.push();
    solver->var2cycle_t.last() = i;
    
    solver->shift_spod_t.push();
    solver->shift_spod_t[i+1] = solver->shift_spod_t[i];

    if (verbose_) printf("[zigzag]  Encoding %d..\n", i);


    list_ands.push_back(unroll_kind_t(i));

    iter_flatten_ptu(i);

    // simple_path_t(i);
    // solver->shift_spod_t[i + 1] += i * dll_latches->sz;
    // assert(solver->shift_spod_t[i + 1] == (solver->shift_spod_t[i] + (i * dll_latches->sz)));


    ands2clauses_t(i, solver->act_trans_t[i]);

    /* PRINT ALL VARS */
    // for (int test = 0; test <= i; test++)
    //   printf("%d ", solver->shift_spod_t[test]);
    // printf(" (SHIFT)\n");
    // for (int test = 0; test <= i; test++)
    //   printf("%d ", solver->act_trans_t[test]);
    // printf(" (ACT_TRANS) | nb_vars : %d\n", solver->nb_vars_by_cycle);

    // idef* tmp_def;
    // dll_elem<idef> *tmp_e;

    // // Inputs
    // tmp_e = dll_inputs->head;
    // while(tmp_e != NULL) {
    //   tmp_def = tmp_e->def;
      
    //   for (int test = 0; test <= i; test++)
    // 	printf("%d ", tmp_def->lhs_ts[test] >> 1);
    //   printf("\n");
    //   tmp_e = tmp_e->next;
    // }  
  
    // // Latches
    // tmp_e = dll_latches->head;
    // while(tmp_e != NULL) {
    //   tmp_def = tmp_e->def;
    //   for (int test = 0; test <= i; test++)
    // 	printf("%d ", tmp_def->lhs_ts[test] >> 1);
    //   printf("\n");
    //   tmp_e = tmp_e->next;
    // }  
    // // Ands
    // tmp_e = dll_ands->head;
    // while(tmp_e != NULL) {
    //   tmp_def = tmp_e->def;
    //   for (int test = 0; test <= i; test++)
    // 	printf("%d ", tmp_def->lhs_ts[test] >> 1);
    //   printf("\n");
    //   tmp_e = tmp_e->next;
    // }  
    /******************/

    if (rep) {
      replicate_clauses_t(i, verbose_);
      if (verbose_) printf("[zigzag]  Internal Replication : %d!\n", solver->stats_replicated_clauses);
    }

    assert(mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] > 1);
    
    int bad = mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i];
    bad = aiger_sign(bad_lit) == aiger_sign(bad) ? (bad >> 1) : -(bad >> 1);
    solver->var2cycle_t[abs(bad)] = -2;

    lbool res_check;
    if (verbose_) printf("[zigzag]  Solving %d-induction..\n", i);
      
    assume(bad); // Assume: -P(x)
      
    for (int k = 0; k <= i; k++)
      assume(solver->act_trans_t[k]); // Assume: T(x, x')

    res_check = check_sat(verbose_);
    if (res_check == l_False) {
      status = 0;
      break;
    } else if (i > 1 && simple == 2) {
      if (simple_path_od_t(i, bad, verbose_)) {
	status = 0;
	break;
      }
    }

    if (verbose_) printf("[zigzag]  Solving BMC %d..\n", i);

    assume(bad);              // Assume: -P(x)
    assume(act_init_t);       // Assume:  I(x)

    for (int k = 0; k <= i; k++)
      assume(solver->act_trans_t[k]); // Assume: T(x, x')

    res_check = check_sat(verbose_);
    if (res_check == l_True) {
      status = 1;
      break;
    }

    printf("u%d\n", i);
    fflush(stdout);

    solver->var2cycle_t[abs(bad)] = i;
    bcl(-bad, -(solver->act_trans_t[i]));                // Add: P(x)
    //ucl(-bad);

  }

  printf("%d\nb0\n.\n", status);
  fflush(stdout);

  if (verbose_)
    printf("Vars %d (%d - %d)| Clauses %d\n", solver->nVars() - solver->eliminated_vars,
	   solver->nVars(), solver->eliminated_vars, solver->nClauses());

  if (verbose_)
    printf("Learnt : %d | STATS REPLICABLE FOUND : %d\n",
	   solver->nLearnts(), solver->stats_replicable_clause_found);

  destroy_model(verbose_);
  return 0;
}

